package owg.lgm;

import owg.util.math.Vec;
import static owg.util.math.Trig.sq;

import owg.util.math.Lerp;
import owg.util.math.Trig;

public class Path
{
	protected boolean closed;
	protected boolean smooth;
	protected int precision;
	protected double[] knots;
	
	protected double[] verts;
	
	protected boolean boxValid = false;
	protected double x0, y0, x1, y1;
	
	public Path(boolean closed, boolean smooth, int precision, double... knots)
	{
		this.closed = closed;
		this.smooth = smooth;
		this.precision = precision;
		this.knots = knots;
		validateVerts();
	}
	
	protected Path(Path src)
	{
		this.closed = src.closed;
		this.smooth = src.smooth;
		this.precision = src.precision;
		this.knots = src.knots;
		this.verts = src.verts;
		this.boxValid = src.boxValid;
		this.x0 = src.x0;
		this.y0 = src.y0;
		this.x1 = src.x1;
		this.y1 = src.y1;
	}
	
	public void validateVerts()
	{
		if(verts == null)
		{
			int count = knots.length/3;
			if(smooth)
			{
				if(closed)
					verts = new double[4*(1+count + (count  )*(2*(precision-1)))];
				else
					verts = new double[4*(  count + (count-2)*(2*(precision-1)))];
				
				int v = closed ? 0 : 4;
				final int n = 1+2*(precision-1);
				
				for(int i = 3; i<knots.length-3; i+=3)
					v += qBezier(verts, v, n, knots, i-3, i, i+3);
				
				if(closed)
				{
					v += qBezier(verts, v, n, knots, knots.length-6, knots.length-3, 0);
					v += qBezier(verts, v, n, knots, knots.length-3, 0, 3);
				}
				else
				{
					verts[0] = knots[0];
					verts[1] = knots[1];
					verts[2] = knots[2];
					verts[verts.length-4] = knots[knots.length-3];
					verts[verts.length-3] = knots[knots.length-2];
					verts[verts.length-2] = knots[knots.length-1];
				}
			}
			else
			{
				verts = new double[4*count + (closed ? 4 : 0)];
				for(int i = 0, j = 0; j<knots.length; i+=4, j+=3)
				{
					verts[i  ] = knots[j  ];
					verts[i+1] = knots[j+1];
					verts[i+2] = knots[j+2];
				}
			}
			
			if(closed)
			{
				verts[verts.length-4] = verts[0];
				verts[verts.length-3] = verts[1];
				verts[verts.length-2] = verts[2];
			}
			
			for(int i = 0, j = 4; j<verts.length-1; i+=4, j+=4)
				verts[j+3] = verts[i+3] + Vec.distance(verts, i, verts, j, 2);
		}
	}
	
	protected void validateBox()
	{
		if(!boxValid)
		{
			validateVerts();
			x0 = Double.MAX_VALUE;
			y0 = Double.MAX_VALUE;
			x1 = -Double.MAX_VALUE;
			y1 = -Double.MAX_VALUE;
			for(int i = 0; i<verts.length; i+=4)
			{
				x0 = Math.min(x0, verts[i  ]);
				x1 = Math.max(x1, verts[i  ]);
				y0 = Math.min(y0, verts[i+1]);
				y1 = Math.max(y1, verts[i+1]);
			}
			boxValid = true;
		}
	}
	
	public void invalidate()
	{
		verts = null;
		boxValid = false;
	}
	
	public double[] getVerts()
	{
		validateVerts();
		return verts;
	}
	
	protected static int qBezier(double[] output, int v, int n, double[] ks, int k0, int k1, int k2)
	{
		final double x1 = ks[k1  ];
		final double y1 = ks[k1+1];
		final double z1 = ks[k1+2];
		final double dx0 = (ks[k0  ]-x1)/2;
		final double dy0 = (ks[k0+1]-y1)/2;
		final double dz0 = (ks[k0+2]-z1)/2;
		final double dx2 = (ks[k2  ]-x1)/2;
		final double dy2 = (ks[k2+1]-y1)/2;
		final double dz2 = (ks[k2+2]-z1)/2;
		
		for(int j = 0; j<n; j++)
		{
			final float ff = (float)(1+j)/(n+1);
			output[v  ] = ( x1 + sq(1-ff)*dx0 + sq(ff)*dx2 );
			output[v+1] = ( y1 + sq(1-ff)*dy0 + sq(ff)*dy2 );
			output[v+2] = ( z1 + sq(1-ff)*dz0 + sq(ff)*dz2 );
			v+=4;
		}
		return 4*n;
	}
	
	@Override
	public Path clone()
	{
		return new Path(this);
	}
	
	public void rotate(double degrees)
	{
		double c = Trig.cos(degrees);
		double s = Trig.sin(degrees);
		
		validateBox();
		double cx = (x0+x1)/2;
		double cy = (y0+y1)/2;
		
		for(int i = 0; i<knots.length; i += 3)
		{
			double x = knots[i  ]-cx;
			double y = knots[i+1]-cy;
			knots[i  ] = cx + x*c + y*s;
			knots[i+1] = cy - x*s + y*c;
		}
		invalidate();
	}
	
	public void shift(double x, double y)
	{
		for(int i = 0; i<knots.length; i += 3)
		{
			knots[i  ] += x;
			knots[i+1] += y;
		}
		invalidate();
	}
	
	public void scale(double x, double y)
	{
		validateBox();
		double cx = (x0+x1)/2;
		double cy = (y0+y1)/2;
		
		for(int i = 0; i<knots.length; i += 3)
		{
			knots[i  ] = cx + (knots[i  ]-cx)*x;
			knots[i+1] = cy + (knots[i+1]-cy)*y;
		}
		invalidate();
	}

	public boolean isClosed()
	{
		return closed;
	}
	public boolean isSmooth()
	{
		return smooth;
	}
	public int getPrecision()
	{
		return precision;
	}

	public static double search(double[] verts, double offset)
	{
		double r = 0;
		for(int i = 7; i<verts.length; i+=4, r++)
		{
			if(offset < verts[i])
			{
				double d0 = verts[i-4];
				double d1 = verts[i  ];
				if(d1 > d0)
				{
					r += (offset-d0)/(d1-d0);
					break;
				}
			}
		}
		return r;
	}
	public static double lerp(double[] verts, double index, int comp)
	{
		int i0 = (int)index;
		if(i0 < index) {
			double frac = index-i0;
			i0 = i0*4+comp;
			return Lerp.lerp(verts[i0], verts[i0+4], frac);
		} else {
			return verts[i0*4+comp];
		}
	}
}
