package owg.lgm.gl;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import owg.util.emptyarray.Empty;
import owg.util.struct.Struct;
import owg.util.struct.Struct.N;

public class BMFont
{
	public static BMFont read(byte[] data) throws IOException
	{
		return read(ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN));
	}
	
	public static BMFont read(ByteBuffer data) throws IOException
	{
		{
			int header;
			if((header=data.getInt()) != 0x03464D42)
				throw new IOException("Expected BMF3, got: "+Integer.toHexString(header));
		}
		BMFont bmFont = new BMFont();
		
		while(data.hasRemaining())
		{
			byte blockType = data.get();
			int blockSize = data.getInt();
			switch(blockType)
			{
				case 1:
					if(bmFont.info != null)
						throw new IOException("Duplicate info block");
					bmFont.info = new Info();
					bmFont.info.fontName = new byte[blockSize-Struct.sizeOf(bmFont.info)];
					Struct.read(bmFont.info, data);
					if(bmFont.info.fontName[bmFont.info.fontName.length-1] != 0)
						throw new IOException("Missing null terminator for font name");
					break;
				case 2:
					if(bmFont.common != null)
						throw new IOException("Duplicate common block");
					bmFont.common = new Common();
					if(blockSize != Struct.sizeOf(bmFont.common))
						throw new IOException("Bad common block size "+blockSize+", expected "+Struct.sizeOf(bmFont.common));
					Struct.read(bmFont.common, data);
					break;
				case 3:
					if(bmFont.pages != null)
						throw new IOException("Duplicate pages block");
					if(blockSize%bmFont.common.pages != 0)
						throw new IOException("Pages block size "+blockSize+" is not a multiple of the page count "+bmFont.common.pages);
					bmFont.pages = new String[bmFont.common.pages];
					byte[] nameBuf = new byte[blockSize / bmFont.common.pages];
					for(int i = 0; i<bmFont.common.pages; i++)
					{
						data.get(nameBuf);
						bmFont.pages[i] = new String(nameBuf, 0, nameBuf.length-1, "UTF-8");
					}
					break;
				case 4:
					if(bmFont.chars != null)
						throw new IOException("Duplicate chars block");
					if(blockSize%Char.SIZE != 0)
						throw new IOException("Chars block size "+blockSize+" is not a multiple of the char struct size "+Char.SIZE);
					bmFont.chars = new Char[blockSize/Char.SIZE];
					for(int i = 0; i<bmFont.chars.length; i++)
					{
						bmFont.chars[i] = new Char();
						Struct.read(bmFont.chars[i], data);
					}
					break;
				case 5:
					if(bmFont.kerningPairs != null)
						throw new IOException("Duplicate kerning pairs block");
					if(blockSize%KerningPair.SIZE != 0)
						throw new IOException("Kerning pairs block size "+blockSize+" is not a multiple of the kerning pair struct size "+KerningPair.SIZE);
					bmFont.kerningPairs = new KerningPair[blockSize/KerningPair.SIZE];
					for(int i = 0; i<bmFont.kerningPairs.length; i++)
					{
						bmFont.kerningPairs[i] = new KerningPair();
						Struct.read(bmFont.kerningPairs[i], data);
					}
					break;
				default:
					throw new IOException("Unknown block type: "+blockType);
			}
		}
		return bmFont;
	}
	
	public static class Info
	{
		@N( 0) public short fontSize;
		@N( 1) public byte bitField;
		@N( 2) public byte charSet;
		@N( 3) public short stretchH;
		@N( 4) public byte aa;
		@N( 5) public byte paddingUp;
		@N( 6) public byte paddingRight;
		@N( 7) public byte paddingDown;
		@N( 8) public byte paddingLeft;
		@N( 9) public byte spacingHoriz;
		@N(10) public byte spacingVert;
		@N(11) public byte outline;
		@N(12) public byte[] fontName = Empty.BYTE_ARRAY;
	}
	
	public static class Common
	{
		@N(0) public short lineHeight;
		@N(1) public short base;
		@N(2) public short scaleW;
		@N(3) public short scaleH;
		@N(4) public short pages;
		@N(5) public byte bitField;
		@N(6) public byte alphaChnl;
		@N(7) public byte redChnl;
		@N(8) public byte greenChnl;
		@N(9) public byte blueChnl;
	}
	
	public static class Char
	{
		public static final int SIZE = 20;
		@N(0) public int id;
		@N(1) public short x;
		@N(2) public short y;
		@N(3) public short width;
		@N(4) public short height;
		@N(5) public short xoffset;
		@N(6) public short yoffset;
		@N(7) public short xadvance;
		@N(8) public byte page;
		@N(9) public byte chnl;
	}
	
	public static class KerningPair
	{
		public static final int SIZE = 10;
		@N(0) public int first;
		@N(1) public int second;
		@N(2) public short amount;
	}
	
	public Info info;
	public Common common;
	public String[] pages;
	public Char[] chars;
	public KerningPair[] kerningPairs;
}
