package owg.lgm.gl;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.WeakHashMap;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2ES1;
import com.jogamp.opengl.util.GLBuffers;

import owg.lgm.IImage;

public class Tex2D extends GLResource
{
	protected static WeakHashMap<GL2ES1, Float> maxTextureMaxAnisotropy = new WeakHashMap<>();
	public static float getMaxTextureMaxAnisotropy(GL2ES1 gl)
	{
		Float r = maxTextureMaxAnisotropy.get(gl);
		if(r != null)
			return r;
		if (gl.getContext().isExtensionAvailable("GL_EXT_texture_filter_anisotropic"))
		{
			float[] gf = new float[1];
			gl.glGetFloatv(GL2ES1.GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, gf, 0);
			r = gf[0];
		}
		else
			r = 1.0f;
		maxTextureMaxAnisotropy.put(gl, r);
		return r;
	}
	
	protected static ByteBuffer makeDirect(byte[] data)
	{
		ByteBuffer r = ByteBuffer.allocateDirect(data.length).order(ByteOrder.nativeOrder());
		r.put(data);
		r.rewind();
		return r;
	}
	
	public final TexFormat format;
	public final ByteBuffer data;
	public final int texWidth, texHeight;
	public final int dataWidth, dataHeight;
	public final boolean mipmaps;
	protected TexSampler sampler;
	
	public float tx0, ty0, tx1, ty1;
	public float mx, my, tw, th;
	
	protected transient GL2ES1 gl;
	protected transient int handle;
	protected transient float currentMaxAnisotropy = 1.0f;
	
	public Tex2D(GL2ES1 gl, IImage image, boolean mipmaps, TexSampler sampler)
	{
		this(gl, image.type.texFormat, makeDirect(image.data), image.width, image.height, mipmaps, sampler);
	}
	
	public Tex2D(GL2ES1 gl, TexFormat format, ByteBuffer data, int width, int height, boolean mipmaps, TexSampler sampler)
	{
		super();
		this.format = format;
		this.data = data;
		this.dataWidth = width;
		this.dataHeight = height;
		this.texWidth = GLBuffers.getNextPowerOf2(width);
		this.texHeight = GLBuffers.getNextPowerOf2(height);
		this.tx0 = 0.5f/texWidth;
		this.ty0 = 0.5f/texHeight;
		this.tx1 = (dataWidth-0.5f)/texWidth;
		this.ty1 = (dataHeight-0.5f)/texHeight;
		this.tw = (dataWidth-1.0f)/texWidth;
		this.th = (dataHeight-1.0f)/texHeight;
		this.mx = ((float)dataWidth)/texWidth;
		this.my = ((float)dataHeight)/texHeight;
		this.mipmaps = mipmaps;
		
		this.sampler = sampler;
		
		init(gl);
	}
	
	protected Tex2D()
	{
		this.format = null;
		this.data = null;
		this.dataWidth = 0;
		this.dataHeight = 0;
		this.texWidth = 0;
		this.texHeight = 0;
		this.tx0 = 0;
		this.ty0 = 0;
		this.tx1 = 0;
		this.ty1 = 0;
		this.tw = 0;
		this.th = 0;
		this.mx = 0;
		this.my = 0;
		this.mipmaps = false;
	}
	
	public TexSampler getSampler()
	{
		return sampler;
	}
	public void setSampler(TexSampler sampler)
	{
		this.sampler = sampler;
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, handle);
		applySampler();
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, 0);
	}
	
	protected void applySampler()
	{
		gl.glTexParameteri(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_TEXTURE_MIN_FILTER, sampler.minFilter.value);
		gl.glTexParameteri(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_TEXTURE_MAG_FILTER, sampler.magFilter.value);
		gl.glTexParameteri(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_TEXTURE_WRAP_S, sampler.wrapS.value);
		gl.glTexParameteri(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_TEXTURE_WRAP_T, sampler.wrapT.value);
		float ani = Math.max(1.0f, Math.min(sampler.maxAnisotropy, getMaxTextureMaxAnisotropy(gl)));
		if(currentMaxAnisotropy != ani)
		{
			//Note: If the extension is not supported, then currentMaxAnisotropy == ani == 1, and this does not run.  
			gl.glTexParameterf(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_TEXTURE_MAX_ANISOTROPY_EXT, ani);
			currentMaxAnisotropy = ani;
		}
	}

	public void update(int y0, int h)
	{
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, handle);
		gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1);
		int p = data.position();
		data.position(y0*dataWidth*format.byteCount);
		gl.glTexSubImage2D(GL2ES1.GL_TEXTURE_2D, 0, 0, y0, dataWidth, h, format.baseFormat, format.dataType.value, data);
		data.position(p);
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, 0);
	}

	@Override
	public int getInitOrder() {
		return 0;
	}

	@Override
	public void init(GL2ES1 gl)
	{
		this.gl = gl;
		{
			int[] r = new int[1];
			gl.glGenTextures(1, r, 0);
			handle = r[0];
		}
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, handle);
		if(mipmaps)
			gl.glTexParameteri(GL2ES1.GL_TEXTURE_2D, GL2ES1.GL_GENERATE_MIPMAP, GL2ES1.GL_TRUE);
		gl.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1);
		if(dataWidth == texWidth && dataHeight == texHeight)
			gl.glTexImage2D(GL2ES1.GL_TEXTURE_2D, 0, format.sizedFormat, texWidth, texHeight, 0, format.baseFormat, format.dataType.value, data);
		else {
			gl.glTexImage2D(GL2ES1.GL_TEXTURE_2D, 0, format.sizedFormat, texWidth, texHeight, 0, format.baseFormat, format.dataType.value, null);
			if(data != null)
				gl.glTexSubImage2D(GL2ES1.GL_TEXTURE_2D, 0, 0, 0, dataWidth, dataHeight, format.baseFormat, format.dataType.value, data);
		}
		applySampler();
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, 0);
	}

	@Override
	public void uninit()
	{
		gl.glDeleteTextures(1, new int[] {handle}, 0);
	}

	public int getHandle()
	{
		return handle;
	}

	public void begin()
	{
		gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, handle);
	}
	public void end()
	{
		//
	}
}
