package owg.lgm.util;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.CharBuffer;
import java.nio.DoubleBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Map.Entry;
import java.util.Set;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.Background;
import owg.lgm.Entity;
import owg.lgm.Fonts;
import owg.lgm.Sprite;
import owg.lgm.gl.BMFontRenderer;
import owg.lgm.gl.GLResource;
import owg.lgm.plaf.Sound;
import owg.util.enumutil.EnumUtil;

public class SaveGame
{
	protected static final int TYPEDEF_FID = -1;
	
	protected static class Typedef {
		Class<?> type;
		public Typedef(Class<?> type) {
			this.type = type;
		}
		@Override
		public int hashCode()
		{
			return ~type.hashCode();
		}
		@Override
		public boolean equals(Object obj) {
			return obj instanceof Typedef && ((Typedef)obj).type.equals(type);
		}
		@Override
		public String toString()
		{
			return "Typedef: "+type.getName();
		}
	}
	
	protected static class FieldMap {
		public final Class<?> type;
		public final ArrayList<Field> list;
		public final HashMap<String, Field> map;
		
		public FieldMap(Class<?> type, boolean write)
		{
			this.type = type;
			if(write) {
				list = new ArrayList<>();
				map = null;
			} else {
				list = null;
				map = new HashMap<>();
			}
			scan(type);
		}
		
		protected void scan(Class<?> type)
		{
			if(type == null)
				return;
			scan(type.getSuperclass());
			for(Field f : type.getDeclaredFields())
			{
				int mods = f.getModifiers();
				if(Modifier.isStatic(mods) || Modifier.isTransient(mods))
					continue;
				f.setAccessible(true);
				if(list != null)
					list.add(f);
				if(map != null)
				{
					String fieldName = f.getName();
					Field r = map.put(fieldName, f);
					if(r != null)
						System.err.println("Warning: Field name collision for "+f.getName()+" for "+this.type+" and "+type);
					
					Map<String, String> rMap = Refactor.classFields.get(type);
					String altName;
					if(rMap != null && (altName=rMap.get(fieldName)) != null)
					{
						map.put(altName, f);
						if(r != null)
							System.err.println("Warning: Alt field name collision for "+f.getName()+" for "+this.type+" and "+type);
					}
					
				}
			}
		}
	}
	
	public interface Restorable {
		public void restore();
	}
	
	public ArrayList<Object> items = new ArrayList<>();
	public HashMap<Object, Integer> ids = new HashMap<>();
	public HashMap<Class<?>, FieldMap> fieldMaps = new HashMap<>();
	
	protected byte[] buf = new byte[4096];
	protected boolean write;
	
	public SaveGame(boolean write)
	{
		this.write = write;
		defineItem(null);
		defineItem(Class.class);
		defineItem(new Typedef(Class.class));
		defineItem(String.class);
		defineItem(new Typedef(String.class));
	}
	
	protected Object defineItem(Object item)
	{
		int id = items.size();
		ids.put(item, id);
		items.add(item);
		return item;
	}

	public void write(DataOutputStream dos, Object item) throws IOException
	{
		Integer id = ids.get(item);
		if(id == null)
		{
			Typedef typedef = new Typedef(item.getClass());
			Integer typeid = ids.get(typedef);
			if(typeid == null)
			{
				writeVarInt(dos, TYPEDEF_FID);
				write(dos, item.getClass());
				defineItem(typedef);
			}
			else
				writeVarInt(dos, typeid);
			
			if(item instanceof Class<?>) {
				write(dos, ((Class<?>)item).getName());
				defineItem(item);
			} else if(item instanceof String) {
				dos.writeUTF((String) item);
				defineItem(item);
			} else if(item instanceof Byte) {
				dos.writeByte((Byte)item);
				defineItem(item);
			} else if(item instanceof Short) {
				dos.writeShort((Short)item);
				defineItem(item);
			} else if(item instanceof Integer) {
				dos.writeInt((Integer)item);
				defineItem(item);
			} else if(item instanceof Long) {
				dos.writeLong((Long)item);
				defineItem(item);
			} else if(item instanceof Float) {
				dos.writeFloat((Float)item);
				defineItem(item);
			} else if(item instanceof Double) {
				dos.writeDouble((Double)item);
				defineItem(item);
			} else if(item instanceof Boolean) {
				dos.writeBoolean((Boolean)item);
				defineItem(item);
			} else if(item instanceof Character) {
				dos.writeChar((Character)item);
				defineItem(item);
			} else if(item instanceof Enum) {
				writeVarInt(dos, ((Enum<?>)item).ordinal());
				defineItem(item);
			} else if(item.getClass().isArray()) {
				defineItem(item);
				int len = Array.getLength(item);
				writeVarInt(dos, len);
				for(int i = 0; i<len; i++)
					write(dos, Array.get(item, i));
			} else if(item instanceof List<?>) {
				defineItem(item);
				List<?> list = (List<?>) item;
				writeVarInt(dos, list.size());
				for(int i = 0; i<list.size(); i++)
					write(dos, list.get(i));
			} else if(item instanceof Set<?>) {
				defineItem(item);
				Set<?> set = (Set<?>) item;
				writeVarInt(dos, set.size());
				for(Object value : set)
					write(dos, value);
			} else if(item instanceof Map<?, ?>) {
				defineItem(item);
				Map<?, ?> map = (Map<?, ?>)item;
				writeVarInt(dos, map.size());
				for(Entry<?, ?> e : map.entrySet()) {
					write(dos, e.getKey());
					write(dos, e.getValue());
				}
			} else if(item instanceof Buffer) {
				defineItem(item);
				Buffer buf = (Buffer)item;
				int capacity = buf.capacity();
				int position = buf.position();
				int limit = buf.limit();
				dos.writeBoolean(buf.isDirect());
				writeVarInt(dos, capacity);
				writeVarInt(dos, position);
				writeVarInt(dos, limit);
				int cc = BufferUtil.sizeOfBufferElem(buf.getClass());
				buf.clear();
				ByteBuffer bytes = ByteBuffer.allocate(buf.capacity() * cc).order(ByteOrder.nativeOrder());
				if(buf instanceof ByteBuffer)
					bytes.put((ByteBuffer)buf);
				else if(buf instanceof ShortBuffer)
					bytes.asShortBuffer().put((ShortBuffer)buf);
				else if(buf instanceof IntBuffer)
					bytes.asIntBuffer().put((IntBuffer)buf);
				else if(buf instanceof LongBuffer)
					bytes.asLongBuffer().put((LongBuffer)buf);
				else if(buf instanceof FloatBuffer)
					bytes.asFloatBuffer().put((FloatBuffer)buf);
				else if(buf instanceof DoubleBuffer)
					bytes.asDoubleBuffer().put((DoubleBuffer)buf);
				else if(buf instanceof CharBuffer)
					bytes.asCharBuffer().put((CharBuffer)buf);
				else
					throw new IllegalArgumentException("Unknown buffer type: "+buf.getClass().getName());
				buf.position(position);
				buf.limit(limit);
				dos.write(bytes.array());
			} else if(item instanceof Sprite) {
				write(dos, ((Sprite) item).name);
				defineItem(item);
			} else if(item instanceof Background) {
				write(dos, ((Background) item).name);
				defineItem(item);
			} else if(item instanceof Sound) {
				write(dos, ((Sound) item).name);
				defineItem(item);
			} else if(item instanceof BMFontRenderer) {
				write(dos, ((BMFontRenderer) item).name);
				defineItem(item);
			} else if(item.getClass().getName().startsWith("owg.")) {
				defineItem(item);
				FieldMap fields = getFieldMap(item.getClass());
				writeVarInt(dos, fields.list.size());
				for(int i = 0; i<fields.list.size(); i++)
				{
					Field f = fields.list.get(i);
					write(dos, f.getName());
					try {
						Object value = f.get(item);
						write(dos, value);
					} catch(IllegalArgumentException | IllegalAccessException err) {
						throw new IllegalArgumentException("Failed to save field: "+item.getClass()+"."+f.getName(), err);
					}
				}
			} else
				throw new IllegalArgumentException("Cannot save object of type: "+item.getClass());
		}
		else
			writeVarInt(dos, id);
	}

	protected FieldMap getFieldMap(Class<?> type)
	{
		FieldMap fields = fieldMaps.get(type);
		if(fields == null) {
			fields = new FieldMap(type, write);
			fieldMaps.put(type, fields);
		}
		return fields;
	}

	protected void writeVarInt(DataOutputStream dos, int value) throws IOException
	{
		while(true)
		{
			int b = value&0x7F;
			value >>>= 7;
			if(value == 0) {
				dos.writeByte(b);
				return;
			} else
				dos.writeByte(b|0x80);
		}
	}
	
	@SuppressWarnings({ "rawtypes", "unchecked" })
	public Object read(DataInputStream dis, Object target) throws IOException
	{
		int id = readVarInt(dis);
		Object ref;
		if(id == TYPEDEF_FID)
		{
			ref = new Typedef((Class<?>)read(dis, null));
			defineItem(ref);
		}
		else
			ref = items.get(id);
		
		if(ref instanceof Typedef)
		{
			//This is an object we haven't seen before
			Class<?> type = ((Typedef) ref).type;
			if(Class.class.isAssignableFrom(type)) {
				String className = (String)read(dis, null);
				if(Refactor.classNames.containsKey(className))
					className = Refactor.classNames.get(className);
				try {
					Class<?> clz = Class.forName(className);
					return defineItem(clz);
				} catch(Exception err) {
					throw new IOException("Failed to read class: "+className, err);
				}
			} else if(String.class.isAssignableFrom(type))
				return defineItem(dis.readUTF());
			else if(Byte.class.isAssignableFrom(type))
				return defineItem(dis.readByte());
			else if(Short.class.isAssignableFrom(type))
				return defineItem(dis.readShort());
			else if(Integer.class.isAssignableFrom(type))
				return defineItem(dis.readInt());
			else if(Long.class.isAssignableFrom(type))
				return defineItem(dis.readLong());
			else if(Float.class.isAssignableFrom(type))
				return defineItem(dis.readFloat());
			else if(Double.class.isAssignableFrom(type))
				return defineItem(dis.readDouble());
			else if(Boolean.class.isAssignableFrom(type))
				return defineItem(dis.readBoolean());
			else if(Character.class.isAssignableFrom(type))
				return defineItem(dis.readChar());
			else if(Enum.class.isAssignableFrom(type))
				return defineItem(EnumUtil.getEnumConstantsImmutable(type).get(readVarInt(dis)));
			else if(type.isArray()) {
				int len = readVarInt(dis);
				Object array;
				if(target != null) {
					array = target;
					int targetLength = Array.getLength(array);
					if(targetLength != len)
						throw new IllegalArgumentException("Final array length mismatch: "+len+" vs "+targetLength);
				} else
					array = Array.newInstance(type.getComponentType(), len);
				defineItem(array);
				for(int i = 0; i<len; i++)
					Array.set(array, i, read(dis, null));
				return array;
			} else if(List.class.isAssignableFrom(type)) {
				int len = readVarInt(dis);
				List list;
				if(target != null) {
					list = (List) target;
					list.clear();
				} else if(ArrayList.class.isAssignableFrom(type))
					list = new ArrayList(len);
				else
					list = (List) getInstance(type);
				defineItem(list);
				for(int i = 0; i<len; i++)
					list.add(read(dis, null));
				return list;
			} else if(Set.class.isAssignableFrom(type)) {
				int len = readVarInt(dis);
				Set set;
				if(target != null) {
					set = (Set) target;
					set.clear();
				} else
					set = (Set) getInstance(type);
				defineItem(set);
				for(int i = 0; i<len; i++)
					set.add(read(dis, null));
				return set;
			} else if(Map.class.isAssignableFrom(type)) {
				int len = readVarInt(dis);
				Map map;
				if(target != null) {
					map = (Map) target;
					map.clear();
				} else
					map = (Map) getInstance(type);
				defineItem(map);
				for(int i = 0; i<len; i++)
					map.put(read(dis, null), read(dis, null));
				return map;
			} else if(Buffer.class.isAssignableFrom(type)) {
				boolean direct = dis.readBoolean();
				int capacity = readVarInt(dis);
				int position = readVarInt(dis);
				int limit = readVarInt(dis);
				int cc = BufferUtil.sizeOfBufferElem((Class<? extends Buffer>) type);
				ByteBuffer bytes;
				if(direct) {
					bytes = ByteBuffer.allocateDirect(capacity * cc).order(ByteOrder.nativeOrder());
					while(bytes.hasRemaining()) {
						int r = dis.read(buf, 0, Math.min(buf.length, bytes.remaining()));
						if(r <= 0)
							throw new EOFException("Unexpected end of file in buffer "+bytes.position()+" / "+capacity);
						bytes.put(buf, 0, r);
					}
				} else {
					bytes = ByteBuffer.allocate(capacity * cc).order(ByteOrder.nativeOrder());
					while(bytes.hasRemaining()) {
						int r = dis.read(bytes.array(), bytes.position(), bytes.remaining());
						if(r <= 0)
							throw new EOFException("Unexpected end of file in buffer "+bytes.position()+" / "+capacity);
						bytes.position(bytes.position()+r);
					}
				}
				bytes.clear();
				
				Buffer buffer;
				if(target != null) {
					buffer = (Buffer) target;
					buffer.clear();
					if(buffer.capacity() != capacity)
						throw new IllegalArgumentException("Final buffer capacity mismatch: "+capacity+" vs "+buffer.capacity());
					if(ByteBuffer.class.isAssignableFrom(type))
						((ByteBuffer)target).put(bytes);
					else if(ShortBuffer.class.isAssignableFrom(type))
						((ShortBuffer)target).put(bytes.asShortBuffer());
					else if(IntBuffer.class.isAssignableFrom(type))
						((IntBuffer)target).put(bytes.asIntBuffer());
					else if(LongBuffer.class.isAssignableFrom(type))
						((LongBuffer)target).put(bytes.asLongBuffer());
					else if(FloatBuffer.class.isAssignableFrom(type))
						((FloatBuffer)target).put(bytes.asFloatBuffer());
					else if(DoubleBuffer.class.isAssignableFrom(type))
						((DoubleBuffer)target).put(bytes.asDoubleBuffer());
					else if(CharBuffer.class.isAssignableFrom(type))
						((CharBuffer)target).put(bytes.asCharBuffer());
					else
						throw new IllegalArgumentException("Unknown buffer type: "+type);
				} else {
					if(ByteBuffer.class.isAssignableFrom(type))
						buffer = bytes;
					else if(ShortBuffer.class.isAssignableFrom(type))
						buffer = bytes.asShortBuffer();
					else if(IntBuffer.class.isAssignableFrom(type))
						buffer = bytes.asIntBuffer();
					else if(LongBuffer.class.isAssignableFrom(type))
						buffer = bytes.asLongBuffer();
					else if(FloatBuffer.class.isAssignableFrom(type))
						buffer = bytes.asFloatBuffer();
					else if(DoubleBuffer.class.isAssignableFrom(type))
						buffer = bytes.asDoubleBuffer();
					else if(CharBuffer.class.isAssignableFrom(type))
						buffer = bytes.asCharBuffer();
					else
						throw new IllegalArgumentException("Unknown buffer type: "+type);
				}
				
				defineItem(buffer);
				buffer.position(position);
				buffer.limit(limit);
				return buffer;
			} else if(Sprite.class.isAssignableFrom(type))
				return defineItem(Sprite.get((String)read(dis, null)));
			else if(Background.class.isAssignableFrom(type))
				return defineItem(Background.get((String)read(dis, null)));
			else if(Sound.class.isAssignableFrom(type))
				return defineItem(Sound.get((String)read(dis, null)));
			else if(BMFontRenderer.class.isAssignableFrom(type))
				return defineItem(Fonts.get((String)read(dis, null)));
			else if(type.getName().startsWith("owg.")) {
				Object obj;
				if(target != null)
					obj = target;
				else
					obj = getInstance(type);
				defineItem(obj);
				int fieldCount = readVarInt(dis);
				FieldMap fields = getFieldMap(type);
				for(int i = 0; i<fieldCount; i++)
				{
					String fieldName = (String)read(dis, null);
					try {
						Field f = fields.map.get(fieldName);
						if(f == null)
							throw new NoSuchFieldException(fieldName);

						int mods = f.getModifiers();
						if(Modifier.isTransient(mods))
							throw new IllegalArgumentException("Cannot read transient field "+type.getName()+"."+fieldName);
						else if(Modifier.isStatic(mods))
							throw new IllegalArgumentException("Cannot read static field "+type.getName()+"."+fieldName);
						else if(!Modifier.isFinal(mods)) {
							f.setAccessible(true);
							Object value = read(dis, null);
							f.set(obj, value);
						} else {
							f.setAccessible(true);
							Object dst = f.get(obj);
							Object value = read(dis, dst);
							if(!Objects.equals(dst, value))
								f.set(obj, value);
							//throw new IllegalArgumentException("Failed to apply final field: "+type.getName()+"."+fieldName+", "+dst+" != "+value);
						}
					} catch(Exception err) {
						throw new IllegalArgumentException("Failed to read field: "+type.getName()+"."+fieldName, err);
					}
				}
				return obj;
			} else
				throw new IllegalArgumentException("Cannot read object of type: "+type);
		}
		else
		{
			//ref is the value
			return ref;
		}
	}

	protected int readVarInt(DataInputStream dis) throws IOException
	{
		int value = 0;
		int shift = 0;
		while(true)
		{
			int r = dis.read();
			if(r == -1)
				throw new EOFException("End of file while reading varint");
			value |= (r&0x7f) << shift;
			if((r&0x80) == 0)
				return value;
			else
				shift += 7;
			if(shift >= 35)
				throw new IOException("Varint is too long!");
		}
	}

	protected Object getInstance(Class<?> type) throws IOException
	{
		//if(type.isMemberClass() && !Modifier.isStatic(type.getModifiers()))
		try {
			if(Entity.class.isAssignableFrom(type))
			{
				Constructor<?> ctor = type.getDeclaredConstructor(double.class, double.class);
				ctor.setAccessible(true);
				return ctor.newInstance(0.0, 0.0);
			}
			else
			{
				Constructor<?> ctor = type.getDeclaredConstructor();
				ctor.setAccessible(true);
				return ctor.newInstance();
			}
		} catch(Exception err) {
			throw new IOException("Failed to spawn object of type: "+type.getName(), err);
		}
	}

	public void restoreAll(GL2ES1 gl)
	{
		for(int i = 0; i<items.size(); i++)
		{
			Object item = items.get(i);
			if(item instanceof Restorable)
				((Restorable) item).restore();
			if(item instanceof GLResource)
				((GLResource) item).init(gl);
		}
	}
}
