package owg.competent;

import java.util.HashMap;

/**Competent Audio (CA) error codes.
 * 
 *All error codes except CA_NO_ERROR are negative, enabling them to be returned from functions that normally return non-negative integers.
 *
 *CA tries to handle exceptional circumstances gracefully as far as possible, so that the application doesn't have to. 
 *For instance, if an audio stream cannot be opened or fails during playback, CA will continue to process events as normal.
 *If the application needs to know, there are dedicated functions to detect such errors after they happen.
 *
 *In theory, most CA functions can return CA_THREAD_ERROR, CA_MUTEX_ERROR, CA_CONDITION_ERROR, or CA_INCONSISTENT_DATA.
 *However, this is not documented on the functions since these errors may only occur in abnormal conditions such as memory corruption.
 *
 *All other error returns are documented on the individual functions.
 *
 *All memory allocations are checked and will return CA_INSUFFICIENT_MEMORY on failure, but attempting to handle this is not recommended.
 *
 *As such, almost all possible error returns are a result of incorrect usage.*/
public enum CAErrorCode
{
	NO_ERROR(0, "No error"),
	//The following error codes are correspond to negative libsoundio errors
	INSUFFICIENT_MEMORY(-1, "Out of memory."),
	ERROR_INIT_AUDIO_BACKEND(-2, "The backend does not appear to be active or running."), 
	ERROR_SYSTEM_RESOURCES(-3, "A system resource other than memory was not available."),
	ERROR_OPENING_DEVICE(-4, "Attempted to open a device and failed."), 
	ERROR_NO_SUCH_DEVICE(-5, "Attempted to open a device that does not exist."),
	ERROR_INVALID(-6, "The programmer did not comply with the API."),
	ERROR_BACKEND_UNAVAILABLE(-7, "libsoundio was compiled without support for that backend."),
	ERROR_STREAMING(-8, "An open stream had an error that can only be recovered from by destroying the stream and creating it again."),
	ERROR_INCOMPATIBLE_DEVICE(-9, "Attempted to use a device with parameters it cannot support."),
	ERROR_NO_SUCH_CLIENT(-10, "JACK returned 'JackNoSuchClient'"),
	ERROR_INCOMPATIBLE_BACKEND(-11, "Attempted to use parameters that the backend cannot support."),
	ERROR_BACKEND_DISCONNECTED(-12, "Backend server shutdown or became inactive."),
	ERROR_INTERRUPTED(-13, "libsoundio was interrupted"),
	ERROR_UNDERFLOW(-14, "Buffer underrun occurred."),
	ERROR_ENCODING_STRING(-15, "Unable to convert to or from UTF-8 to the native string format."),
	
	//Note: The following 2 error codes exist only in libsoundio-ca
	ERROR_MISSING_SHARED_OBJECT(-50, "The shared object file for the backend could not be loaded."),
	ERROR_MISSING_SYMBOLS(-51, "The backend dynamic library was loaded but was missing symbols."),

	//CA original codes 
	ALREADY_INITIALIZED(-100, "Cannot initialize when Competent Audio is already initialized"), 
	NOT_INITIALIZED("Competent Audio is not initialized"),
	BACKEND_NOT_INITIALIZED("No attempt has been made to initialize the specified backend"),

	THREAD_ERROR("Thread-related error"),
	MUTEX_ERROR("Mutex-related error"),
	CONDITION_ERROR("Condition variable-related error"),
	
	INCONSISTENT_DATA("Inconsistent data detected due to internal error or forbidden usage"),
	
	NULL_POINTER("Illegal NULL pointer passed to function"),
	INDEX_OUT_OF_BOUNDS("Index is out of bounds"),

	ERROR_SAMPLE_FORMAT("Bad, unsupported or invalid sample format"),
	ERROR_SPEAKER_LAYOUT("Bad, unsupported or invalid speaker layout"),
	ERROR_SAMPLE_RATE("Bad, unsupported or invalid sample rate"),

	NOT_A_SOURCE("Specified machine is not a source machine"),
	NOT_A_DESTINATION("Specified machine is not a destination machine"),
	WRONG_MACHINE_TYPE("Function requires a specific machine type that differs from the type of the given machine"),
	WRONG_DESTINATION_FUNC("Destination object needs a specific pull function to use this function"),

	INVALID_FREQUENCY("Frequency value must be positive"),
	INVALID_GAIN("Gain value must be finite"),
	INVALID_CHUNK_SIZE("Chunk size must be positive"),
	BAD_CHANNEL_VECTOR_SIZE("Channel vector size does not match the sound model"),

	DSP_TOO_LARGE("DSP too large: 4*frameSize*maxChunkSize + stackSize + progLen must not exceed 0xFFFF");
	  

	private static class ErrorMap
	{
		private static int lastCode;
		private static HashMap<Integer, CAErrorCode> map = new HashMap<Integer, CAErrorCode>();
	}

	/**Returns the error enum corresponding to the given error code, or null if the code does not correspond to any known error.*/
	public static CAErrorCode getErrorOrNull(int code)
	{
		return ErrorMap.map.get(code);
	}
	/**Returns the error enum corresponding to the given error code, 
	 * or throws a {@link CAException} with the code {@link #INCONSISTENT_DATA} if the code does not correspond to any known error.*/
	public static CAErrorCode getError(int code) throws CAException
	{
		CAErrorCode result = ErrorMap.map.get(code);
		if(result == null)
			throw new CAException(INCONSISTENT_DATA, "code="+code);
		return result;
	}
	/**The integer error code.*/
	public final int code;
	/**A human-readable english description of the error.*/
	public final String description;

	private CAErrorCode(String description)
	{
		this(ErrorMap.lastCode+1, description);
	}

	private CAErrorCode(int code, String description)
	{
		ErrorMap.lastCode = code;
		ErrorMap.map.put(code, this);
		this.code = code;
		this.description = description;
	}
	
	@Override
	public String toString() {
		return code+": "+description;
	}
}
