package owg.competent;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.Entry;

import owg.util.nativelibraryloader.NativeLibraryLoader;

/**A specialized {@link NativeLibraryLoader} for Competent Audio native library files.
 * The convenience function {@link #initialize()} will automatically handle everything for most applications.*/
public class CALoader extends NativeLibraryLoader {
	protected static final PlatformSupport[] supports;
	static
	{
		//Note: This creates a support declaration for all 4 of these architectures
		//for all of linux, windows and mac os, but most of these are just theoretical.
		//There are no plans to actually attempt to compile jcompetent_audio for all
		//of them.
		final HashMap<String, HashSet<String>> archs = new HashMap<String, HashSet<String>>();
		archs.put("x86", X86);
		archs.put("amd64", AMD64);
		archs.put("arm", ARM);
		archs.put("aarch64", AARCH64);		
		final String version = "_"+CANatives.JCOMPETENT_AUDIO_VERSION;
		supports = new PlatformSupport[3*archs.size()];
		
		int i = 0;
		for(Entry<String, HashSet<String>> e : archs.entrySet())
		{
			final String archName = e.getKey();
			final HashSet<String> supportArch = e.getValue();
			
			supports[i++] = new PlatformSupport(new PlatformMatcher() {
				@Override
				public boolean match(String osName, String osArch) {
					return isLinux(osName) && supportArch.contains(osArch);
				}
			}, "linux_"+archName+version, "libsoundio-ca.so.2", "libcompetent_audio.so", "libjcompetent_audio.so");
			
			supports[i++] = new PlatformSupport(new PlatformMatcher() {
				@Override
				public boolean match(String osName, String osArch) {
					return isWindows(osName) && supportArch.contains(osArch);
				}
			}, "win_"+archName+version, "libsoundio-ca.dll", "competent_audio.dll", "jcompetent_audio.dll");
			
			supports[i++] = new PlatformSupport(new PlatformMatcher() {
				@Override
				public boolean match(String osName, String osArch) {
					return isMacOs(osName) && supportArch.contains(osArch);
				}
			}, "macos_"+archName+version, "libsoundio-ca.dylib", "libcompetent_audio.dylib", "libjcompetent_audio.dylib");
		}
		assert i == supports.length;
	}
	
	/**Creates a {@link CALoader} which will extract to and/or load the native library files from the given directory.*/
	public CALoader(File libraryDir) {
		super(CALoader.class, libraryDir, true, supports);
	}
	
	/**<p>
	 * Extracts the Competent Audio native libraries to a temporary directory, loads them into the JVM, calls {@link CompetentAudio#initialize()}, and installs a shutdown hook to invoke
	 * {@link CompetentAudio#uninitialize()} when the JVM shuts down.
	 * </p><p>
	 * If {@link CompetentAudio} is already initialized, this function does nothing.
	 * </p>*/
	public synchronized static void initialize() throws CAException
	{
		synchronized (CompetentAudio.initLock) {
			if(CompetentAudio.initialized)
				return;
			
			try {
				String tmpDir = System.getProperty("java.io.tmpdir");
				if(tmpDir == null)
					throw new IOException("Could not locate temporary directory: java.io.tmpdir is null");
				File dir = new File(tmpDir, "jcompetent_audio");
				CALoader loader = new CALoader(dir);
				loader.extract();
				loader.load();
			} catch(IOException err) {
				throw new CAException(CAErrorCode.ERROR_MISSING_SHARED_OBJECT, null, err);
			} catch(UnsatisfiedLinkError err) {
				throw new CAException(CAErrorCode.ERROR_MISSING_SYMBOLS, null, err);
			}
			
			CompetentAudio.CA.initialize();
			Runtime.getRuntime().addShutdownHook(new Thread(){
				@Override
				public void run() {
					CompetentAudio.CA.uninitialize();
				}
			});			
		}
	}
}
